from django.http import HttpResponse
from django.views.generic import ListView, DetailView, CreateView, UpdateView
from django.contrib.auth.mixins import PermissionRequiredMixin
from rest_framework import viewsets
from rest_framework import permissions
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import generics
from . import models, serializers


######################################################
##  Project Views
######################################################

class ProjectListView(PermissionRequiredMixin, ListView):
    model = models.Project
    paginate_by = 50
    permission_required = "crunch.view_project"


class ProjectDetailView(PermissionRequiredMixin, DetailView):
    model = models.Project
    permission_required = "crunch.view_project"
    # lookup_field = 'slug'


class ProjectCreateView(PermissionRequiredMixin, CreateView):
    model = models.Project
    permission_required = "crunch.add_project"


class ProjectUpdateView(PermissionRequiredMixin, UpdateView):
    model = models.Project
    template_name = "crunch/form.html"
    # form_class = ProjectForm
    permission_required = "crunch.update_project"
    extra_context = dict(
        form_title="Update Project",
    )


class ProjectAPI(viewsets.ModelViewSet):
    """
    API endpoint that allows projects to be viewed or edited.
    """
    queryset = models.Project.objects.all()
    serializer_class = serializers.ProjectSerializer
    permission_classes = [permissions.DjangoModelPermissions]
    lookup_field = 'slug'


######################################################
##  Dataset Views
######################################################


class DatasetDetailView(PermissionRequiredMixin, DetailView):
    model = models.Dataset
    permission_required = "crunch.view_dataset"
    lookup_field = 'slug'


class DatasetAPI(viewsets.ModelViewSet):
    """
    API endpoint that allows datasets to be viewed or edited.
    """
    queryset = models.Dataset.objects.all()
    serializer_class = serializers.DatasetSerializer
    permission_classes = [permissions.DjangoModelPermissions]
    lookup_field = 'slug'


class DatasetCreateView(PermissionRequiredMixin, CreateView):
    model = models.Dataset
    permission_required = "crunch.add_dataset"


class DatasetUpdateView(PermissionRequiredMixin, UpdateView):
    model = models.Dataset
    template_name = "crunch/form.html"
    # form_class = DatasetForm
    permission_required = "crunch.update_dataset"
    extra_context = dict(
        form_title="Update Dataset",
    )


class ProjectNextDatasetReference(APIView):
    """
    Retuns the study accession ID and the batch index to process next for a particular project.
    """
    permission_classes = [permissions.IsAuthenticated] # should be 'view_dataset'
        
    def get(self, request, format=None, slug=None):
        assert slug is not None
        project = models.Project.objects.get(slug=slug)
        dataset = project.next_unprocessed_dataset()

        dataset_reference = dict(project=dataset.parent.slug, dataset=dataset.slug) if dataset else dict(project="", dataset="")
        serializer = serializers.DatasetReferenceSerializer(dataset_reference)
        return Response(serializer.data)


class NextDatasetReference(APIView):
    """
    Retuns the study accession ID and the batch index to process next.
    """
    permission_classes = [permissions.IsAuthenticated] # should be 'view_dataset'
        
    def get(self, request, format=None):
        dataset = models.Dataset.next_unprocessed()

        dataset_reference = dict(project=dataset.parent.slug, dataset=dataset.slug) if dataset else dict(project="", dataset="")
        serializer = serializers.DatasetReferenceSerializer(dataset_reference)
        return Response(serializer.data)


class StatusListCreateAPIView(generics.ListCreateAPIView):
    queryset = models.Status.objects.all()
    serializer_class = serializers.StatusSerializer

    def perform_create(self, serializer):
        serializer.save(
            site_user=self.request.user,
        )


######################################################
##  Item Views
######################################################

class ItemDetailView(PermissionRequiredMixin, DetailView):
    model = models.Item
    permission_required = "crunch.view_item"
    lookup_field = 'slug'


class ItemCreateView(PermissionRequiredMixin, CreateView):
    model = models.Item
    permission_required = "crunch.add_item"


class ItemUpdateView(PermissionRequiredMixin, UpdateView):
    model = models.Item
    template_name = "crunch/form.html"
    # form_class = ItemForm
    permission_required = "crunch.update_item"
    extra_context = dict(
        form_title="Update Item",
    )


class ItemMapView(ItemDetailView):
    def get(self, request, slug) -> HttpResponse:
        item = self.get_object()
        map = item.map()
        html = map.to_html(as_string=True) if map else "<p>No map available</p>"
        return HttpResponse(html)


class ItemAPI(viewsets.ModelViewSet):
    """
    API endpoint that allows items to be viewed or edited.
    """
    queryset = models.Item.objects.all()
    serializer_class = serializers.ItemSerializer
    permission_classes = [permissions.DjangoModelPermissions]
    lookup_field = 'slug'


######################################################
##  Attribute Views
######################################################

class CharAttributeAPI(viewsets.ModelViewSet):
    queryset = models.CharAttribute.objects.all()
    serializer_class = serializers.CharAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class FloatAttributeAPI(viewsets.ModelViewSet):
    queryset = models.FloatAttribute.objects.all()
    serializer_class = serializers.FloatAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class IntegerAttributeAPI(viewsets.ModelViewSet):
    queryset = models.IntegerAttribute.objects.all()
    serializer_class = serializers.IntegerAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class FilesizeAttributeAPI(viewsets.ModelViewSet):
    queryset = models.FilesizeAttribute.objects.all()
    serializer_class = serializers.FilesizeAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class BooleanAttributeAPI(viewsets.ModelViewSet):
    queryset = models.BooleanAttribute.objects.all()
    serializer_class = serializers.BooleanAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class FloatAttributeAPI(viewsets.ModelViewSet):
    queryset = models.FloatAttribute.objects.all()
    serializer_class = serializers.FloatAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class URLAttributeAPI(viewsets.ModelViewSet):
    queryset = models.URLAttribute.objects.all()
    serializer_class = serializers.URLAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class LatLongAttributeAPI(viewsets.ModelViewSet):
    queryset = models.LatLongAttribute.objects.all()
    serializer_class = serializers.LatLongAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class DateTimeAttributeAPI(viewsets.ModelViewSet):
    queryset = models.DateTimeAttribute.objects.all()
    serializer_class = serializers.DateTimeAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]


class DateAttributeAPI(viewsets.ModelViewSet):
    queryset = models.DateAttribute.objects.all()
    serializer_class = serializers.DateAttributeSerializer
    permission_classes = [permissions.DjangoModelPermissions]



from rest_framework import serializers

from . import models

class ProjectSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Project
        fields = ['id', 'name', 'slug', 'description', 'details', 'workflow']


class AttributeSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Attribute
        fields = ['id', 'key', ]

    def to_representation(self, instance):
        return instance.value_dict()


class AbstractAttributeSerializer(serializers.ModelSerializer):
    item = serializers.SlugRelatedField(slug_field='slug', queryset=models.Item.objects.all())
    class Meta:
        fields = [
            "item",
            "key",
            "value",
        ]


class CharAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.CharAttribute


class FloatAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.FloatAttribute


class IntegerAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.IntegerAttribute


class FilesizeAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.FilesizeAttribute


class BooleanAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.BooleanAttribute


class URLAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.URLAttribute


class DateTimeAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.DateTimeAttribute


class DateAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.DateAttribute


class LatLongAttributeSerializer(AbstractAttributeSerializer):
    class Meta(AbstractAttributeSerializer.Meta):
        model = models.LatLongAttribute
        fields = [
            "item",
            "key",
            "latitude",
            "longitude",
        ]


class ItemSerializer(serializers.HyperlinkedModelSerializer):
    parent = serializers.SlugRelatedField(slug_field='slug', queryset=models.Item.objects.all())
    attributes = AttributeSerializer(many=True, required=False)

    class Meta:
        model = models.Item
        fields = ['id', 'name', 'slug','parent', 'description', 'details', 'attributes']


class DatasetSerializer(serializers.HyperlinkedModelSerializer):
    parent = serializers.SlugRelatedField(slug_field='slug', queryset=models.Project.objects.all())
    attributes = AttributeSerializer(many=True, required=False)
    items = ItemSerializer(many=True, required=False)

    class Meta:
        model = models.Dataset
        fields = ['id', 'name', 'slug','parent', 'description', 'details', 'attributes', 'items', 'base_file_path']


class DatasetReferenceSerializer(serializers.Serializer):
    project = serializers.CharField(max_length=255)
    dataset = serializers.CharField(max_length=255)

    
class StatusSerializer(serializers.ModelSerializer):
    dataset = serializers.PrimaryKeyRelatedField(queryset=models.Dataset.objects.all())

    class Meta:
        model = models.Status
        fields = [
            "id", 
            "dataset", 
            "site_user", 
            "stage", 
            "state", 
            "note",
            "agent_user",
            "version",
            "revision",
            "system",
            "system_release",
            "system_version",
            "machine",
            "hostname",
            "ip_address",
            "mac_address",
            "memory_total",
            "memory_free",
            "disk_total",
            "disk_free",
        ]




# Generated by Django 3.2.12 on 2022-08-11 01:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crunch', '0006_filesizeattribute'),
    ]

    operations = [
        migrations.AddField(
            model_name='dataset',
            name='base_file_path',
            field=models.CharField(default='', max_length=4096),
            preserve_default=False,
        ),
    ]



from typing import Dict, Union
from datetime import datetime
import requests
from functools import cached_property
from pathlib import Path
import traceback
import subprocess
import json
from django.core.files.storage import DefaultStorage

from crunch.django.app.enums import Stage, State
from crunch.django.app import storages
from rich.console import Console

console = Console()

from . import utils
from .connections import Connection
from .enums import WorkflowType, RunResult

STAGE_STYLE = "bold red"

class Run():
    """
    An object to manage processing a crunch dataset.
    """
    def __init__(
        self, 
        connection:Connection, 
        dataset_slug:str, 
        storage_settings:Union[Dict,Path], 
        working_directory:Path, 
        workflow_type:WorkflowType, 
        workflow_path:Path=None, 
        cores:str="1",
    ):
        self.connection = connection
        self.dataset_slug = dataset_slug

        if not self.dataset_slug:
            raise ValueError("Please specifiy dataset.")

        self.dataset_data = connection.get_json_response(f"/api/datasets/{dataset_slug}/")
        self.workflow_type = workflow_type
        self.workflow_path = workflow_path
        self.cores = cores
        self.storage_settings = storage_settings
        self.setup_md5_checksums = dict()

        working_directory = Path(working_directory)
        working_directory.mkdir(exist_ok=True, parents=True)
        self.working_directory = working_directory
        
        # TODO raise exception
        assert self.dataset_data["slug"] == dataset_slug
        self.project = self.dataset_data["parent"]
        self.dataset_id = self.dataset_data["id"]
        self.base_file_path = self.dataset_data["base_file_path"]

    def send_status(self, state, note:str="") -> requests.Response:
        """ Sends a status update about the processing of this dataset. """
        return self.connection.send_status(
            self.dataset_id, 
            stage=self.current_stage, 
            state=state, 
            note=note,
        )

    @cached_property
    def crunch_subdir(self) -> Path:
        """ Returns the path to the .crunch subdirectory in the working directory for this dataset. """
        crunch_subdir = self.working_directory/".crunch"        
        crunch_subdir.mkdir(exist_ok=True, parents=True)
        return crunch_subdir

    @cached_property
    def storage(self) -> DefaultStorage:
        """ Gets the default storage object. """
        return storages.get_storage_with_settings(self.storage_settings)

    def setup(self) -> RunResult:
        """ 
        Sets up this dataset for processing.
        
        This involves:

        - Copying the initial data from storage
        - Saving the MD5 checksums for all the initial data in ``.crunch/setup_md5_checksums.json``
        - Saves the metadata for the dataset in ``.crunch/dataset.json``
        - Saves the metadata for the project in ``.crunch/project.json``
        - Creates the script to run the workflow (either a bash script or a Snakefile for Snakemake)


        Returns:
            RunResult: Whether or not this stage was successful.
        """
        self.current_stage = Stage.SETUP
        console.print(f"Setup stage {self.dataset_slug}", style=STAGE_STYLE)
        try:
            self.send_status(State.START)
            
            # Pull data from storage
            storages.copy_recursive_from_storage(
                self.base_file_path, self.working_directory, storage=self.storage
            )
            self.setup_md5_checksums = utils.md5_checksums(self.working_directory)
            with open(self.crunch_subdir / "setup_md5_checksums.json", "w", encoding="utf-8") as f:
                json.dump(self.setup_md5_checksums, f, ensure_ascii=False, indent=4)

            # TODO check to see if dataset.json already exists
            with open(self.crunch_subdir / "dataset.json", "w", encoding="utf-8") as f:
                json.dump(self.dataset_data, f, ensure_ascii=False, indent=4)

            # get project details
            project_data = self.connection.get_json_response(f"/api/projects/{self.project}/")
            # TODO raise exception
            assert project_data["slug"] == self.project

            # TODO check to see if project.json already exists
            with open(self.crunch_subdir / "project.json", "w", encoding="utf-8") as f:
                json.dump(project_data, f, ensure_ascii=False, indent=4)

            # get snakefile or script
            if not self.workflow_path:
                self.workflow_path = utils.write_workflow(
                    project_data["workflow"], 
                    working_directory=self.crunch_subdir,
                    workflow_type=self.workflow_type,
                )

            self.send_status(State.SUCCESS)
        except Exception as e:
            console.print(f"Setup failed {self.dataset_slug}: {e}", style=STAGE_STYLE)
            self.send_status(State.FAIL, note=str(e))
            return RunResult.FAIL
        
        return RunResult.SUCCESS

    def workflow(self) -> RunResult:
        """ 
        Runs the workflow on a dataset that has been set up.
        
        This involves running a bash script as a subprocess or running Snakemake with a Snakefile.

        Returns:
            RunResult: Whether or not this stage was successful.
        """
        self.current_stage = Stage.WORKFLOW

        try:
            self.send_status(State.START)
            if self.workflow_type == WorkflowType.snakemake:
                import snakemake

                args = [
                    f"--snakefile={self.workflow_path}",
                    "--use-conda",
                    f"--cores={self.cores}",
                    f"--directory={self.working_directory}",
                    f"--conda-frontend={utils.conda_frontend()}"
                ]

                try:
                    snakemake.main(args)
                except SystemExit as result:
                    print(f"result {result}")
            elif self.workflow_type == WorkflowType.script:
                subprocess.run(f"{self.workflow_path.resolve()}", capture_output=True, check=True, cwd=self.working_directory)

            self.send_status(State.SUCCESS)
        except Exception as e:
            console.print(f"Workflow failed {self.dataset_slug}: {e}", style=STAGE_STYLE)
            self.send_status(State.FAIL, note=str(e))
            return RunResult.FAIL
        
        return RunResult.SUCCESS

    def upload(self) -> RunResult:
        """
        Uploads new or modified files to the storage for the dataset.

        It also creates the following files:
        - .crunch/upload_md5_checksums.json which lists all MD5 checksums after the dataset has finished.
        - .crunch/deleted.txt which lists all files that were present after setup but which were deleted as the workflow ran.

        Returns:
            RunResult: Whether or not this stage was successful.
        """
        self.current_stage = Stage.UPLOAD
        console.print(f"Upload stage {self.dataset_slug}", style=STAGE_STYLE)
        try:
            self.send_status(State.START)

            # calculate checksums
            self.upload_md5_checksums = utils.md5_checksums(self.working_directory)
            upload_md5_checksums_path = self.crunch_subdir / "upload_md5_checksums.json"
            with open(upload_md5_checksums_path, "w", encoding="utf-8") as f:
                json.dump(self.upload_md5_checksums, f, ensure_ascii=False, indent=4)

            setup_files = set(self.setup_md5_checksums.keys())
            upload_files = set(self.upload_md5_checksums.keys())
            new_files = upload_files - setup_files
            deleted_files = setup_files - upload_files
            remaining_files = setup_files.intersection(upload_files)
            modified_files = set(
                file for file in remaining_files 
                if self.upload_md5_checksums[file] != self.setup_md5_checksums[file]
            )
            deleted_log_path = self.crunch_subdir / "deleted.txt"
            deleted_log_path.write_text("\n".join(deleted_files))

            files_to_upload = modified_files | new_files | set([upload_md5_checksums_path, deleted_log_path])
            paths_to_upload = [self.working_directory/file for file in files_to_upload]

            storages.copy_to_storage(
                paths_to_upload, 
                local_dir=self.working_directory,
                base=self.base_file_path, 
                storage=self.storage,
            )

            # TODO Option to delete on remote storage?

            self.send_status(State.SUCCESS)
        except Exception as e:
            console.print(f"Upload failed {self.dataset_slug}: {e}", style=STAGE_STYLE)
            traceback.print_exc()
            self.send_status(State.FAIL, note=str(e))
            return RunResult.FAIL
        
        return RunResult.SUCCESS

    def __call__(self) -> RunResult:
        console.print(f"Processing '{self.dataset_slug}'.")
        self.setup_result = self.setup()
        if self.setup_result:
            return self.setup_result

        self.workflow_result = self.workflow()
        if self.workflow_result:
            return self.workflow_result

        self.upload_result = self.upload()
        return self.upload_result




# Generated by Django 3.2.15 on 2022-09-08 04:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crunch', '0010_auto_20220908_0001'),
    ]

    operations = [
        migrations.AddField(
            model_name='dataset',
            name='locked',
            field=models.BooleanField(default=False, help_text='If the dataset is locked then it will not show up in the loop of available datasets.'),
        ),
    ]



# Generated by Django 3.2.12 on 2022-03-20 06:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('crunch', '0002_auto_20220320_0535'),
    ]

    operations = [
        migrations.CreateModel(
            name='BooleanAttribute',
            fields=[
                ('attribute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.attribute')),
                ('value', models.BooleanField()),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.attribute',),
        ),
    ]



from django.contrib import admin
from polymorphic.admin import PolymorphicParentModelAdmin, PolymorphicChildModelAdmin, PolymorphicChildModelFilter

from . import models

class ProjectChildAdmin(PolymorphicChildModelAdmin):
    """ Base admin class for all child models """
    base_model = models.Project  # Optional, explicitly set here.


# @admin.register(models.Project)
# class ProjectParentAdmin(PolymorphicParentModelAdmin):
#     """ The parent model admin """
#     base_model = models.Project  # Optional, explicitly set here.
#     list_filter = (PolymorphicChildModelFilter,)  # This is optional.


class AttributeChildAdmin(PolymorphicChildModelAdmin):
    base_model = models.Attribute


@admin.register(models.CharAttribute)
class CharAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.IntegerAttribute)
class IntegerAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.FilesizeAttribute)
class FilesizeAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.FloatAttribute)
class FloatAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.URLAttribute)
class URLAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.LatLongAttribute)
class LatLongAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.DateTimeAttribute)
class DateTimeAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.DateAttribute)
class DateAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.BooleanAttribute)
class BooleanAttributeAdmin(AttributeChildAdmin):
    pass


@admin.register(models.Attribute)
class AttributeParentAdmin(PolymorphicParentModelAdmin):
    base_model = models.Attribute  # Optional, explicitly set here.
    child_models = (
        models.IntegerAttribute, 
        models.FilesizeAttribute, 
        models.FloatAttribute, 
        models.CharAttribute,
        models.URLAttribute,
        models.BooleanAttribute,
        models.DateAttribute,
        models.DateTimeAttribute,
    ) # This should be automatic
    list_filter = (PolymorphicChildModelFilter,)  # This is optional.


admin.site.register(models.Project)
admin.site.register(models.Dataset)
admin.site.register(models.Item)
admin.site.register(models.Status)
# admin.site.register(models.Attribute)





import pydeck
import pandas as pd


def item_map(item):
    geo_attributes = item.descendant_latlongattributes()

    plot_data = [
        dict(
            latitude=float(attribute.latitude),
            longitude=float(attribute.longitude),
            item=str(attribute.item),
            url=str(attribute.item.get_absolute_url()),
        )
        for attribute in geo_attributes
    ]
    df = pd.DataFrame(plot_data)

    # ICON_URL = "https://upload.wikimedia.org/wikipedia/commons/9/92/Ic_location_on_48px.svg"
    # ICON_WIDTH = 48
    # ICON_HEIGHT = 48

    ICON_URL = "https://upload.wikimedia.org/wikipedia/commons/6/65/OOjs_UI_icon_mapPin-progressive.svg"
    ICON_WIDTH = 20
    ICON_HEIGHT = 20

    icon_data = {
        "url": ICON_URL,
        "width": ICON_WIDTH,
        "height": ICON_HEIGHT,
        "anchorY": ICON_HEIGHT,
    }

    df["icon_data"] = None
    for i in df.index:
        df.at[i, "icon_data"] = icon_data

    layer = pydeck.Layer(
        "IconLayer",
        df,
        pickable=True,
        get_icon="icon_data",
        get_size=20,
        get_position=["longitude", "latitude"],
    )

    view_state = pydeck.ViewState(
        latitude=df["latitude"].mean(),
        longitude=df["longitude"].mean(),
        zoom=2,
        min_zoom=1,
        max_zoom=16,
        pitch=0,
        bearing=0,
    )

    map = pydeck.Deck(layers=[layer], initial_view_state=view_state, map_style="road")

    return map



from os import getenv
from typing import Union, Dict
from unicodedata import decimal
import requests
from rest_framework import status as drf_status
import enum
from datetime import datetime, date
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from rich.console import Console
from crunch.django.app.enums import Stage, State

console = Console()

from . import diagnostics

CRUNCH_URL_KEY = "CRUNCH_URL"
CRUNCH_TOKEN_KEY = "CRUNCH_TOKEN"

class CrunchAPIException(Exception):
    """ Raised when there is an error getting information from the API of a crunch site. """
    pass


class Connection():
    """
    An object to manage calls to the REST API of a crunch hosted site.
    """
    def __init__(self, base_url:str = None, token:str = None, verbose:bool = False):
        """
        An object to manage calls to the REST API of a crunch hosted site.

        Args:
            base_url (str, optional): The URL for the endpoint for the project on the crunch hosted site. 
                If not provided then it attempts to use the 'CRUNCH_URL' environment variable.
            token (str, optional): An access token for a user on the crunch hosted site. 
                If not provided then it attempts to use the 'CRUNCH_TOKEN' environment variable.

        Raises:
            CrunchAPIException: If the `base_url` is not provided and it is not available using the 'CRUNCH_URL' environment variable.
            CrunchAPIException: If the `token` is not provided and it is not available using the 'CRUNCH_TOKEN' environment variable.
        """
        self.base_url = base_url or getenv(CRUNCH_URL_KEY, None)
        if not self.base_url:
            raise CrunchAPIException(f"Please provide a base URL to a crunch hosted site. This can be set using the '{CRUNCH_URL_KEY}' environment variable.")

        self.token = token or getenv(CRUNCH_TOKEN_KEY, None)
        if not self.token:
            raise CrunchAPIException(f"Please provide an authentication token. This can be set using the '{CRUNCH_TOKEN_KEY}' environment variable.")

        self.verbose = verbose

    def get_headers(self) -> dict:
        """
        Creates the headers needed to API calls to the REST API on a crunch hosted site.

        Used internally when making GET and POST requests using this class.

        Raises:
            CrunchAPIException: Raised if no valid token is available.

        Returns:
            dict: The headers for API calls as a Python dictionary.
        """
        headers = {"Authorization": f"Token {self.token}" }
        return headers

    def absolute_url(self, relative_url:str) -> str:
        if self.base_url.endswith("/"):
            self.base_url = self.base_url[:-1]
        if relative_url.startswith("/"):
            relative_url = relative_url[1:]

        url = f"{self.base_url}/{relative_url}"
        return url

    def post(self, relative_url:str, **kwargs) -> requests.Response:
        url = self.absolute_url(relative_url) 
        if self.verbose:
            console.print(f"Posting to {url} : {kwargs}")

        result = requests.post(url, headers=self.get_headers(), data=kwargs)
        if self.verbose:
            console.print(f"Response {result.status_code}: {result.reason}")

        if result.status_code >= 400:
            console.print(f"Failed posting to {url} : {kwargs}")
            console.print(result.json())
        return result

    def add_project(self, project:str, description:str="", details:str="") -> requests.Response:
        """
        Creates a new project on a hosted django-crunch site.

        Args:
            project (str): The name of the new crunch project.
            description (str, optional): A brief description of this new project. Defaults to "".
            details (str, optional): A long description of this project in Markdown format. Defaults to "".

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        if self.verbose:
            console.print(f"Adding project '{project}' on the site {self.base_url}")

        return self.post(
            "api/projects/", 
            name=project,
            description=description,
            details=details,
        )

    def add_dataset(self, project:str, dataset:str, description:str="", details:str="") -> requests.Response:
        """
        Creates a new dataset and adds it to a project on a hosted django-crunch site.

        Args:
            project (str): The slug of the project that this dataset is to be added to.
            dataset (str): The name of the new dataset.
            description (str, optional): A brief description of this new dataset. Defaults to "".
            details (str, optional): A long description of this dataset in Markdown format. Defaults to "".

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        if self.verbose:
            console.print(f"Adding dataset '{dataset}' to project '{project}' on the site {self.base_url}")

        return self.post(
            "api/datasets/", 
            parent=project,
            name=dataset,
            description=description,
            details=details,
        )

    def add_item(self, parent:str, item:str, description:str="", details:str="") -> requests.Response:
        """
        Creates a new item on a hosted django-crunch site.

        Args:
            parent (str): The slug of the parent item that this item is to be added to.
            item (str): The name of the new item.
            description (str, optional): A brief description of this new dataset. Defaults to "".
            details (str, optional): A long description of this dataset in Markdown format. Defaults to "".

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        if self.verbose:
            console.print(f"Adding item '{item}' to parent '{parent}' on the site {self.base_url}")

        return self.post(
            "api/items/", 
            parent=parent,
            name=item,
            description=description,
            details=details,
        )

    def add_key_value_attribute(self, url:str, item:str, key:str, value) -> requests.Response:
        """
        Adds an attribute as a key/value pair on an item. 
        
        This is mainly used by other methods on this class to add attributes with specific types.

        Args:
            url (str): The relative URL for adding this type of attribute on the crunch site. For this, see urls.py in the crunch Django app.
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value: The data to be used for this attribute. The object needs to be serializable.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        if self.verbose:
            console.print(f"Adding attribute '{key}'->'{value}' to item '{item}' on the hosted site {self.base_url}")

        return self.post(
            url, 
            item=item,
            key=key,
            value=value,
        )

    def add_char_attribute(self, item:str, key:str, value:str) -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is a string of characters. 

        Args:
            project (str): The slug for the project.
            dataset (str): The slug for the dataset.
            key (str): The key for this attribute.
            value (str): The string of characters for this attribute.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        return self.add_key_value_attribute(
            url="api/attributes/char/", 
            item=item,
            key=key,
            value=value,
        ) 

    def add_attributes(self, item:str, **kwargs):
        """
        Adds multiple attributes as a key/value pairs on a dataset. 

        Each type is inferred from the type of the value.

        Args:
            item (str): The slug for the item.
            **kwargs: key/value pairs to add as char attributes.
        """
        url_validator = URLValidator()
        for key, value in kwargs.items():
            if isinstance(value, str):
                try: 
                    url_validator(value)
                    self.add_url_attribute(item=item, key=key, value=value)
                except ValidationError:
                    self.add_char_attribute(item=item, key=key, value=value)
            elif isinstance(value, float):
                self.add_float_attribute(item=item, key=key, value=value)
            elif isinstance(value, bool): # This needs to be before the 'int' one because bools are also integers in python
                self.add_boolean_attribute(item=item, key=key, value=value)
            elif isinstance(value, int):
                self.add_integer_attribute(item=item, key=key, value=value)
            elif isinstance(value, datetime):
                self.add_datetime_attribute(item=item, key=key, value=value)
            elif isinstance(value, date):
                self.add_date_attribute(item=item, key=key, value=value)
            else:
                raise CrunchAPIException(f"Cannot infer type of value '{value}' ({type(value).__name__}). (The key was '{key}')")

    def add_float_attribute(self, item:str, key:str, value:float) -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is a float. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value (str): The float value for this attribute.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        return self.add_key_value_attribute(
            url="api/attributes/float/", 
            item=item,
            key=key,
            value=value,
        )                

    def add_datetime_attribute(self, item:str, key:str, value:Union[datetime,str], format:str="") -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is a datetime. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value (Union[datetime,str]): The value for this attribute as a datetime or a string.
            format (str): If the `value` is a string then this format string can be used with datetime.strptime to convert to a datetime object. If no format is given then the string is interpreted using dateutil.parser.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        if isinstance(value, str):
            if format:
                value = datetime.strptime(value, format)
            else:
                from dateutil import parser
                value = parser.parse(value)  

        return self.add_key_value_attribute(
            url="api/attributes/datetime/", 
            item=item,
            key=key,
            value=value,
        )                

    def add_date_attribute(self, item:str, key:str, value:Union[date,str], format:str="") -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is a datetime. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value (Union[date,str]): The value for this attribute as a date or a string.
            format (str): If the `value` is a string then this format string can be used with datetime.strptime to convert to a date object. If no format is given then the string is interpreted using dateutil.parser.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        if isinstance(value, str):
            if format:
                value = datetime.strptime(value, format)
            else:
                from dateutil import parser
                value = parser.parse(value)  

        if isinstance(value, datetime):
            value = value.date()

        return self.add_key_value_attribute(
            url="api/attributes/date/", 
            item=item,
            key=key,
            value=value,
        )                

    def add_integer_attribute(self, item:str, key:str, value:int) -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is an integer. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value (int): The integer value for this attribute.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        return self.add_key_value_attribute(
            url="api/attributes/int/", 
            item=item,
            key=key,
            value=value,
        )               
        
    def add_filesize_attribute(self, item:str, key:str, value:int) -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is an filesize. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value (int): The size of the file in bytes.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        return self.add_key_value_attribute(
            url="api/attributes/filesize/", 
            item=item,
            key=key,
            value=value,
        )               
                
    def add_boolean_attribute(self, item:str, key:str, value:bool) -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is a boolean. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value (bool): The integer value for this attribute.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        return self.add_key_value_attribute(
            url="api/attributes/bool/", 
            item=item,
            key=key,
            value=value,
        )                  

    def add_url_attribute(self, item:str, key:str, value:str) -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is a URL. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            value (str): The str value for this attribute.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        return self.add_key_value_attribute(
            url="api/attributes/url/", 
            item=item,
            key=key,
            value=value,
        )       

    def add_lat_long_attribute(self, item:str, key:str, latitude:Union[str,float,decimal], longitude:Union[str,float,decimal]) -> requests.Response:
        """
        Adds an attribute as a key/value pair on a dataset when the value is a coordinate with latitude and longitude. 

        Args:
            item (str): The slug for the item.
            key (str): The key for this attribute.
            latitude (Union[str,float,decimal]): The latitude for this coordinate.
            longitude (Union[str,float,decimal]): The longitude for this coordinate.

        Returns:
            requests.Response: The request object from posting to the crunch API.
        """
        if self.verbose:
            console.print(f"Adding attribute '{key}'->'{latitude},{longitude}' to item '{item}' on the hosted site {self.base_url}")

        return self.post(
            relative_url="api/attributes/lat-long/",
            item=item,
            key=key,
            latitude=latitude,
            longitude=longitude,
        )

    def send_status(self, dataset_id:str, stage:Stage, state:State, note:str="") -> requests.Response:
        """
        Sends an update of the status of one stage in processing a dataset.

        Args:
            dataset_id (str): The ID of the dataset for this status update.
            stage (Stage): The stage of this status update.
            state (State): The state of this status update.
            note (str, optional): A note which gives more information to this status update. Defaults to "".

        Raises:
            CrunchAPIException: If there was an error posting this status update to the API.

        Returns:
            requests.Response: The resulting response from the request to the API.
        """
        data = dict(
            dataset=dataset_id,
            stage=stage.value,
            state=state.value,
            note=note,
        )
        data.update( diagnostics.get_diagnostics() )
        result = self.post("api/statuses/", **data)
        if result.status_code >= 400:
            raise CrunchAPIException(f"Failed sending status.\n{result.status_code}: {result.reason}\nData: {data}")

        return result

    def get_request(self, relative_url:str):
        url = self.absolute_url(relative_url)
        return requests.get(url, headers=self.get_headers())
        
    def get_json_response( self, relative_url:str ) -> Dict:
        """
        Requests JSON data from the API of a crunch hosted site and returns it as a dictionary.

        Args:
            relative_url (str): The URL path relative to the base URL of the endpoint for the project on the crunch hosted site.

        Raises:
            CrunchAPIException: Raises exception if there is an error getting a JSON response from the API.

        Returns:
            Dict: The JSON data from the API encoded as a dictionary.
        """
        response = self.get_request(relative_url)
        json_response = response.json()
        
        if len(json_response.keys()) == 1 and "detail" in json_response:
            raise CrunchAPIException(
                f"Error getting JSON response from URL '{self.absolute_url(relative_url)}':\n" +
                f"{json_response['detail']}"
            )

        return json_response



from django.core.management.base import BaseCommand, CommandError
from crunch.django.app.models import Dataset

class Command(BaseCommand):
    help = 'Lists the files in storage associated with a dataset.'

    def add_arguments(self, parser):
        parser.add_argument('project', type=str)
        parser.add_argument('dataset', type=str)

    def handle(self, *args, **options):
        dataset = Dataset.objects.get(slug=options['dataset'], project__slug=options['project'])
        print(dataset)
        print(list(dataset.files()))
        


# Generated by Django 3.2.12 on 2022-03-20 05:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('crunch', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='attribute',
            name='dataset',
        ),
        migrations.AddField(
            model_name='attribute',
            name='item',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='attributes', to='crunch.item'),
            preserve_default=False,
        ),
    ]



import logging
from typing import Dict, Union
import toml
import json
from operator import mod
from pathlib import Path
from django.core.files import File
from django.conf import settings
from anytree import NodeMixin, RenderTree, PreOrderIter
from django.core.files.storage import default_storage, DefaultStorage
import os
import time
import shutil
import datetime


class Directory:
    pass


def get_storage_with_settings(storage_settings:Union[Dict,Path]) -> DefaultStorage:
    """
    Configures the Django config with settings if it has not been configured yet and then returns the default storage.

    Useful for connecting to a Django storage object outside of a webserver.

    Args:
        storage_settings (Union[Dict,Path]): The settings with which to configure the Django config if it has not been configured already.
            Can be a dictionary with settings or can be a path to a toml or json file with the settings.

    Raises:
        IOError: Raised if passed a Path object which isn't a toml or json file.
        ValueError: Raised if the settings cannot be interpreted as a dictionary.

    Returns:
        DefaultStorage: The default storage object used by Django.
    """
    if not settings.configured:
        if isinstance(storage_settings, Path):
            with open(storage_settings) as storage_settings_file:
                suffix = storage_settings.suffix.lower()
                if suffix == ".toml":
                    storage_settings = toml.load(storage_settings_file)
                elif suffix == ".json":
                    storage_settings = json.load(storage_settings_file)
                else:
                    raise IOError(f"Cannot find interpreter for {storage_settings}")

        if not isinstance(storage_settings, dict):
            raise ValueError(
                f"Storage settings of type {type(storage_settings)} unable to be read. " +
                "Please pass a dictionary or a path to a toml or json file."
            )
        settings.configure(**storage_settings)
    return default_storage


class StorageDirectory(NodeMixin):
    def __init__(
        self, *args, base_path: Union[str,Path], storage=None, parent=None, children=None, **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.base_path = Path(base_path)
        self.storage = storage
        self.parent = parent
        if children:
            self.children = children

    def __str__(self):
        return str(self.base_path)

    def __repr__(self):
        return str(self)

    def files(self):
        """Returns a list of StorageFile objects for files in this immediate directory."""
        return [child for child in self.children if isinstance(child, StorageFile)]

    def short_str(self):
        if self.parent:
            return str(self.base_path.relative_to(self.parent.base_path))
        return ""

    def render(self):
        result = f"{self.base_path}\n"
        for pre, _, node in RenderTree(self):
            if node == self:
                continue
            treestr = f"{pre}{node.short_str()}"
            result += treestr.ljust(8) + "\n"
        return result

    def render_html(self):
        try:
            result = "<div>"
            result += f"{self.base_path}<br>\n"
            for pre, _, node in RenderTree(self):
                if node == self:
                    continue

                if isinstance(node, StorageFile):
                    result += f"{pre}<a href='{node.url()}'>{node.short_str()}</a><br>\n"
                else:
                    result += f"{pre}{node.short_str()}<br>\n"
            result += "</div>"
        except Exception as err:
            result = f"<div>Failed to read storage at {self.base_path}</div>"
            
        return result

    def directory_descendents(self, stop=None, maxlevel: int = None):
        """Does a pre order iteration of subdirectories."""
        return PreOrderIter(
            self,
            filter_=lambda node: isinstance(node, StorageDirectory),
            stop=stop,
            maxlevel=maxlevel,
        )

    def file_descendents(self, stop=None, maxlevel: int = None):
        """Does a pre order iteration of subdirectories and returns the files in them."""
        return PreOrderIter(
            self,
            filter_=lambda node: isinstance(node, StorageFile),
            stop=stop,
            maxlevel=maxlevel,
        )


class StorageFile(NodeMixin):
    def __init__(self, *args, filename, parent, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = filename
        self.parent = parent

    def __str__(self):
        return self.filename

    def short_str(self):
        return str(self)

    def __repr__(self):
        return str(self)

    def path(self) -> Path:
        return Path(self.parent.base_path, self.filename)

    def url(self):
        storage = self.parent.storage or default_storage
        return storage.url(str(self.path()))


def storage_walk(
    base_path="/", storage=None, parent=None
) -> StorageDirectory:
    """
    Recursively walks a folder, using Django's File Storage.

    Adapted from https://gist.github.com/dvf/c103e697dab77c304d39d60cf279c500
    """
    if storage is None:
        storage = default_storage

    folders, filenames = storage.listdir(str(base_path))

    directory = StorageDirectory(base_path=base_path, parent=parent, storage=storage)

    for subfolder in folders:
        # On S3, we don't have subfolders, so exclude "."
        if subfolder == ".":
            continue

        new_base = Path(base_path, subfolder)
        storage_walk(
            base_path=new_base,
            storage=storage,
            parent=directory,
        )

    for filename in filenames:
        StorageFile(filename=filename, parent=directory)

    return directory


def default_dataset_path(project_slug, dataset_slug):
    return Path("crunch", project_slug, dataset_slug)


def copy_recursive_to_storage(local_dir=".", base="/", storage=None):    
    copy_to_storage(local_dir.rglob("*"), local_dir=local_dir, base=base, storage=storage)


def copy_to_storage(paths, local_dir, base="/", storage=None):
    base = Path(base)
    if storage is None:
        storage = default_storage

    for local_path in paths:
        if local_path.is_dir():
            continue

        local_relative_path = local_path.relative_to(local_dir)
        remote_path = str(base / local_relative_path)

        print(
            f"Copying '{local_path}' from local directory '{local_dir}' to storage at '{remote_path}'"
        )
        with local_path.open(mode="rb") as f:
            storage._save(remote_path, File(f, name=str(local_path)))  


def copy_recursive_from_storage(base="/", local_dir=".", storage=None):
    base = Path(base)
    local_dir = Path(local_dir)
    if storage is None:
        storage = default_storage

    dir_object = storage_walk(base_path=base, storage=storage)
    subdirs = dir_object.directory_descendents()

    for subdir in subdirs:
        listing_path = Path(subdir.base_path)
        relative_path = listing_path.relative_to(base)
        local_path = local_dir / relative_path
        local_path.mkdir(exist_ok=True, parents=True)

        for file in subdir.files():
            filename = file.filename
            print(
                f"Copying '{filename}' in '{listing_path}' from storage to '{local_path}'"
            )
            with storage.open(str(listing_path / filename), "rb") as source:
                with open(local_path / filename, "wb") as target:
                    shutil.copyfileobj(source, target, length=1024)



# Generated by Django 3.2.15 on 2022-09-08 00:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crunch', '0009_auto_20220811_0231'),
    ]

    operations = [
        migrations.AlterField(
            model_name='status',
            name='disk_free',
            field=models.BigIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True),
        ),
        migrations.AlterField(
            model_name='status',
            name='disk_total',
            field=models.BigIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True),
        ),
        migrations.AlterField(
            model_name='status',
            name='memory_free',
            field=models.BigIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True),
        ),
        migrations.AlterField(
            model_name='status',
            name='memory_total',
            field=models.BigIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True),
        ),
    ]



from django.urls import include, path
from django.urls import reverse_lazy
from django.views.generic import RedirectView
from rest_framework import routers

from . import views

router = routers.DefaultRouter()
router.register(r'projects', views.ProjectAPI)
router.register(r'datasets', views.DatasetAPI)
router.register(r'items', views.ItemAPI)
router.register(r'attributes/char', views.CharAttributeAPI)
router.register(r'attributes/float', views.FloatAttributeAPI)
router.register(r'attributes/int', views.IntegerAttributeAPI)
router.register(r'attributes/filesize', views.FilesizeAttributeAPI)
router.register(r'attributes/bool', views.BooleanAttributeAPI)
router.register(r'attributes/url', views.URLAttributeAPI)
router.register(r'attributes/lat-long', views.LatLongAttributeAPI)
router.register(r'attributes/datetime', views.DateTimeAttributeAPI)
router.register(r'attributes/date', views.DateAttributeAPI)

app_name = "crunch"
urlpatterns = [
    path('', views.ProjectListView.as_view(), name='project-list'),
    path('api/', include( (router.urls, 'api') )),
    path('api/statuses/', views.StatusListCreateAPIView.as_view(), name='status-list'),
    path('api/next/', views.NextDatasetReference.as_view(), name='next'),

    path('projects/', RedirectView.as_view(url="..", permanent=False)),
    path("projects/create/", views.ProjectCreateView.as_view(), name="project-create"),
    path('projects/<str:slug>/', views.ProjectDetailView.as_view(), name='project-detail'),
    path("projects/<str:slug>/update/", views.ProjectUpdateView.as_view(), name="project-update"),
    path("api/projects/<str:slug>/next/", views.ProjectNextDatasetReference.as_view(), name="project-api-next"),
    
    path("datasets/create/", views.DatasetCreateView.as_view(), name="dataset-create"),
    path('projects/<str:project>/datasets/', RedirectView.as_view(url="..", permanent=False)),
    path('projects/<str:project>/datasets/<str:slug>', views.DatasetDetailView.as_view(), name='dataset-detail'),
    path("projects/<str:project>/datasets/<str:slug>/update/", views.DatasetUpdateView.as_view(), name="dataset-update"),

    path("items/create/", views.ItemCreateView.as_view(), name="item-create"),
    path('items/<str:slug>/', views.ItemDetailView.as_view(), name='item-detail'),
    path('items/<str:slug>/map/', views.ItemMapView.as_view(), name='item-map'),
    path('items/<str:slug>/update/', views.ItemUpdateView.as_view(), name='item-update'),
]


import stat
import subprocess
from pathlib import Path
import hashlib

from .enums import WorkflowType


def has_mamba()->bool:
    """
    Checks to see if mamba is available.

    Returns:
        bool: Whether or not mamba is in the path.
    """
    try:
        subprocess.run(["mamba", "--version"], capture_output=True, check=True)
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False

    return True


def conda_frontend() -> str:
    return "mamba" if has_mamba() else "conda"


def write_workflow(data:str, working_directory:Path, workflow_type:WorkflowType) -> Path:
    assert data
    working_directory = Path(working_directory)
    working_directory.mkdir(exist_ok=True, parents=True)
    
    if workflow_type == WorkflowType.snakemake:
        workflow_path = working_directory / "Snakefile"
    elif workflow_type == WorkflowType.script:
        workflow_path = working_directory / "script.sh"
    
    with open(workflow_path, "w", encoding="utf-8") as f:
        f.write(data.replace('\r\n', '\n'))

    if workflow_type == WorkflowType.script:
        workflow_path.chmod(workflow_path.stat().st_mode | stat.S_IEXEC)

    return workflow_path


def md5_checksums(directory):
    directory = Path(directory)
    result = dict()
    for path in directory.rglob("*"):
        if path.is_dir():
            continue
        
        relative_path = path.relative_to(directory)
        md5 = hashlib.md5(path.read_bytes()).hexdigest()
        result[str(relative_path)] = md5

    return result



from pathlib import Path
import typer
from rich.console import Console

console = Console()

from crunch.django.app import storages
from . import connections
from .diagnostics import get_diagnostics
from .enums import WorkflowType
from .run import Run


class NoDatasets(Exception):
    """Raised when there are no more datasets to process on a crunch site."""
    pass


app = typer.Typer()

url_arg = typer.Option(
    ...,
    envvar="CRUNCH_URL",
    help="The URL for the endpoint for the project on the crunch hosted site.",
    prompt=True,
)
token_arg = typer.Option(
    ...,
    envvar="CRUNCH_TOKEN",
    help="An access token for a user on the hosted site.",
    prompt=True,
)
dataset_slug_arg = typer.Argument(..., help="The slug for the dataset.")
item_slug_arg = typer.Argument(..., help="The slug for the item.")
storage_settings_arg = typer.Option(
    ..., 
    envvar="CRUNCH_STORAGE_SETTINGS",
    help="The path to a JSON or TOML file with the Django settings for the storage.",
    prompt=True,
)
key_arg = typer.Argument(..., help="The key for this attribute.")
value_arg = typer.Argument(..., help="The value of this attribute.")
directory_arg = typer.Option(
    "./tmp", help="The path to a directory to store the temporary files."
)
cores_arg = typer.Option(
    "1",
    help="The maximum number of cores/jobs to use to run the workflow. If 'all' then it uses all available cores.",
)
workflow_type_arg = typer.Option("snakemake", help="Workflow type (snakemake/script).")
path_arg = typer.Option(None, help="The path to the workflow file.")


@app.command()
def run(
    dataset: str = dataset_slug_arg,
    storage_settings: Path = storage_settings_arg,
    directory: Path = directory_arg,
    cores: str = cores_arg,
    url: str = url_arg,
    token: str = token_arg,
    workflow: WorkflowType = workflow_type_arg,
    path: Path = path_arg,
):
    """
    Processes a dataset.
    """
    r = Run(
        connection=connections.Connection(url, token), 
        dataset_slug=dataset, 
        working_directory=Path(directory),
        workflow_type=workflow, 
        storage_settings=storage_settings,
        workflow_path=path, 
        cores=cores,
    )

    r()


@app.command()
def next(
    storage_settings: Path = storage_settings_arg,
    directory: Path = directory_arg,
    cores: str = cores_arg,
    url: str = url_arg,
    token: str = token_arg,
    project: str = typer.Option(
        "",
        help="The slug for a project the dataset is in. If not given, then it chooses any project.",
    ),
    workflow: WorkflowType = workflow_type_arg,
    path: Path = path_arg,
):
    """
    Processes the next dataset in a project.
    """
    console.print(f"Processing the next dataset from {url}")

    connection = connections.Connection(url, token)

    next_url = f"api/projects/{project}/next/" if project else "api/next/"
    next = connection.get_json_response(next_url)

    if "dataset" in next and "project" in next and next["project"] and next["dataset"]:
        run(
            dataset=next["dataset"],
            storage_settings=storage_settings,
            url=url,
            token=token,
            directory=directory,
            cores=cores,
            workflow=workflow,
            path=path,
        )
    else:
        console.print("No more datasets to process.")
        raise NoDatasets


@app.command()
def loop(
    storage_settings: Path = storage_settings_arg,
    directory: Path = directory_arg,
    cores: str = cores_arg,
    url: str = url_arg,
    token: str = token_arg,
    project: str = typer.Option(
        "",
        help="The slug for a project the dataset is in. If not given, then it chooses any project.",
    ),
    workflow: WorkflowType = workflow_type_arg,
    path: Path = path_arg,
):
    """
    Loops through all the datasets in a project and stops when complete.
    """
    console.print(
        f"Looping through all the datasets from {url} and will stop when complete."
    )
    while True:
        try:
            next(
                url=url,
                token=token,
                storage_settings=storage_settings,
                directory=directory,
                cores=cores,
                workflow=workflow,
                path=path,
                project=project,
            )
        except NoDatasets:
            console.print("Loop concluded.")
            break


@app.command()
def add_project(
    project: str = typer.Argument(..., help="The name of the new crunch project."),
    description: str = typer.Option(
        "", help="A brief description of this new project."
    ),
    details: str = typer.Option(
        "", help="A long description of this project in Markdown format."
    ),
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new project on the hosted site.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_project(
        project=project, description=description, details=details
    )


@app.command()
def add_dataset(
    project: str = typer.Argument(
        ..., help="The slug of the project that this dataset is to be added to."
    ),
    dataset: str = typer.Argument(..., help="The name of the new dataset"),
    description: str = typer.Option(
        "", help="A brief description of this new dataset."
    ),
    details: str = typer.Option(
        "", help="A long description of this dataset in Markdown format."
    ),
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new dataset to a project on the hosted site.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_dataset(
        project=project, dataset=dataset, description=description, details=details
    )


@app.command()
def add_item(
    parent: str = typer.Argument(
        ..., help="The slug of the item that this item is to be added to."
    ),
    item: str = typer.Argument(..., help="The name of the new item"),
    description: str = typer.Option("", help="A brief description of this new item."),
    details: str = typer.Option(
        "", help="A long description of this item in Markdown format."
    ),
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new item to an item on the hosted site.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_item(
        parent=parent, item=item, description=description, details=details
    )


@app.command()
def add_char_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: str = value_arg,
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new char attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_char_attribute(item=item, key=key, value=value)


@app.command()
def add_float_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: float = value_arg,
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new float attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_float_attribute(item=item, key=key, value=value)


@app.command()
def add_datetime_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: str = value_arg,
    format: str = "",
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new datetime attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_datetime_attribute(
        item=item, key=key, value=value, format=format
    )


@app.command()
def add_date_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: str = value_arg,
    format: str = "",
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new date attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_date_attribute(item=item, key=key, value=value, format=format)


@app.command()
def add_lat_long_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    latitude: str = typer.Argument(
        ..., help="The latitude of the coordinate in decimal degrees."
    ),
    longitude: str = typer.Argument(
        ..., help="The longitude of the coordinate in decimal degrees."
    ),
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new lat-long coorinate as an attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_lat_long_attribute(
        item=item, key=key, latitude=latitude, longitude=longitude
    )


@app.command()
def add_integer_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: int = value_arg,
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new integer attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_integer_attribute(item=item, key=key, value=value)


@app.command()
def add_filesize_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: int = value_arg,
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new filesize attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_filesize_attribute(item=item, key=key, value=value)


@app.command()
def add_boolean_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: bool = value_arg,
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new boolean attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_boolean_attribute(item=item, key=key, value=value)


@app.command()
def add_url_attribute(
    item: str = item_slug_arg,
    key: str = key_arg,
    value: str = value_arg,
    verbose: bool = True,
    url: str = url_arg,
    token: str = token_arg,
):
    """
    Adds a new URL attribute to an item.
    """
    connection = connections.Connection(url, token, verbose=verbose)
    return connection.add_url_attribute(item=item, key=key, value=value)


@app.command()
def diagnostics():
    """Display system diagnostics."""
    console.print(get_diagnostics())


@app.command()
def files(
    dataset: str = dataset_slug_arg,
    storage_settings: Path = storage_settings_arg,
    url: str = url_arg,
    token: str = token_arg,
):
    """Displays the files for a dataset."""
    connection = connections.Connection(url, token)
    dataset_data = connection.get_json_response(f"/api/datasets/{dataset}/")
    base_file_path = dataset_data.get("base_file_path")

    storage = storages.get_storage_with_settings(storage_settings)
    listing = storages.storage_walk(base_file_path, storage=storage)
    console.print(listing.render())



# Generated by Django 3.2.12 on 2022-03-20 23:12

from django.db import migrations
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('crunch', '0004_dateattribute'),
    ]

    operations = [
        migrations.AlterField(
            model_name='item',
            name='slug',
            field=django_extensions.db.fields.AutoSlugField(blank=True, editable=False, max_length=255, populate_from='name', unique=True),
        ),
    ]



from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch import receiver
from rest_framework.authtoken.models import Token

@receiver(post_save, sender=settings.AUTH_USER_MODEL)
def create_auth_token(sender, instance=None, created=False, **kwargs):
    if created:
        Token.objects.create(user=instance)



from classytags.helpers import InclusionTag
from classytags.arguments import Argument
from classytags.core import Options
from django import template

register = template.Library()


class ItemMapTag(InclusionTag):
    name = "item_map"
    template = "crunch/item_map.html"
    options = Options(
        Argument("item"),
    )

    def get_context(self, context, item):
        return {"item": item}


register.tag(ItemMapTag)


class DatasetListTag(InclusionTag):
    name = "dataset_list"
    template = "crunch/dataset_list.html"
    options = Options(
        Argument("datasets"),
    )

    def get_context(self, context, datasets):
        return {"datasets": datasets}


register.tag(DatasetListTag)



from enum import Enum


class WorkflowType(str, Enum):
    snakemake = "snakemake"
    script = "script"


class RunResult(Enum):
    SUCCESS = 0
    FAIL = 1

    def __bool__(self):
        return self.value > 0



from django.apps import AppConfig

class CrunchConfig(AppConfig):
    name = "crunch"

    def ready(self):
        super().ready()
        
        # needed because the autodiscover doesn't work unless admin.py is in top directory of app
        from . import admin, tokens



# Generated by Django 3.2.12 on 2022-03-20 04:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields
import polymorphic_tree.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('key', models.CharField(max_length=255)),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_crunch.attribute_set+', to='contenttypes.contenttype')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=1023, unique=True)),
                ('description', models.CharField(blank=True, default='', help_text='A short description in a sentence or more of this item.', max_length=1023)),
                ('details', models.TextField(blank=True, default='', help_text='A detailed description of this item (written in Markdown).')),
                ('slug', django_extensions.db.fields.AutoSlugField(blank=True, editable=False, populate_from='name', unique=True)),
                ('lft', models.PositiveIntegerField(editable=False)),
                ('rght', models.PositiveIntegerField(editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(editable=False)),
                ('parent', polymorphic_tree.models.PolymorphicTreeForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_DEFAULT, related_name='children', to='crunch.item')),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_crunch.item_set+', to='contenttypes.contenttype')),
            ],
            options={
                'ordering': ('tree_id', 'lft'),
                'abstract': False,
                'base_manager_name': 'objects',
                'unique_together': {('parent', 'slug')},
            },
        ),
        migrations.CreateModel(
            name='CharAttribute',
            fields=[
                ('attribute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.attribute')),
                ('value', models.CharField(max_length=1023)),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.attribute',),
        ),
        migrations.CreateModel(
            name='Dataset',
            fields=[
                ('item_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.item')),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.item',),
        ),
        migrations.CreateModel(
            name='DateTimeAttribute',
            fields=[
                ('attribute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.attribute')),
                ('value', models.DateTimeField()),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.attribute',),
        ),
        migrations.CreateModel(
            name='FloatAttribute',
            fields=[
                ('attribute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.attribute')),
                ('value', models.FloatField()),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.attribute',),
        ),
        migrations.CreateModel(
            name='IntegerAttribute',
            fields=[
                ('attribute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.attribute')),
                ('value', models.IntegerField()),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.attribute',),
        ),
        migrations.CreateModel(
            name='LatLongAttribute',
            fields=[
                ('attribute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.attribute')),
                ('latitude', models.DecimalField(decimal_places=9, help_text='The latitude of this location in decimal degrees.', max_digits=12)),
                ('longitude', models.DecimalField(decimal_places=9, help_text='The longitude of this location in decimal degrees.', max_digits=12)),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.attribute',),
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('item_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.item')),
                ('workflow', models.TextField(blank=True, default='', help_text='URL to snakemake repository or text of snakefile.')),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.item',),
        ),
        migrations.CreateModel(
            name='URLAttribute',
            fields=[
                ('attribute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='crunch.attribute')),
                ('value', models.URLField(max_length=1023)),
            ],
            options={
                'abstract': False,
            },
            bases=('crunch.attribute',),
        ),
        migrations.CreateModel(
            name='Status',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('stage', models.IntegerField(choices=[(1, 'Setup'), (2, 'Workflow'), (3, 'Upload')])),
                ('state', models.IntegerField(choices=[(1, 'Start'), (2, 'Success'), (3, 'Fail')])),
                ('note', models.TextField(blank=True, default='')),
                ('agent_user', models.CharField(blank=True, default='', help_text='The name of the user running the agent (see https://docs.python.org/3/library/getpass.html).', max_length=255)),
                ('version', models.CharField(blank=True, default='', help_text='The django-crunch version number of the agent.', max_length=255)),
                ('revision', models.CharField(blank=True, default='', help_text='The django-crunch git revision hash of the agent.', max_length=255)),
                ('system', models.CharField(blank=True, default='', help_text="Returns the system/OS name, such as 'Linux', 'Darwin', 'Java', 'Windows' (see https://docs.python.org/3/library/platform.html).", max_length=255)),
                ('system_release', models.CharField(blank=True, default='', help_text="Returns the system’s release, e.g. '2.2.0' or 'NT' (see https://docs.python.org/3/library/platform.html).", max_length=255)),
                ('system_version', models.CharField(blank=True, default='', help_text="Returns the system’s release version, e.g. '#3 on degas' (see https://docs.python.org/3/library/platform.html).", max_length=255)),
                ('machine', models.CharField(blank=True, default='', help_text="Returns the machine type, e.g. 'i386' (see https://docs.python.org/3/library/platform.html).", max_length=255)),
                ('hostname', models.CharField(blank=True, default='', help_text='The hostname of the machine where the agent was running (see https://docs.python.org/3/library/socket.html).', max_length=255)),
                ('ip_address', models.CharField(blank=True, default='', help_text='The hostname in IPv4 address format (see https://docs.python.org/3/library/socket.html).', max_length=255)),
                ('mac_address', models.CharField(blank=True, default='', help_text='The hardware address  (see https://docs.python.org/3/library/uuid.html).', max_length=255)),
                ('memory_total', models.PositiveIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True)),
                ('memory_free', models.PositiveIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True)),
                ('disk_total', models.PositiveIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True)),
                ('disk_free', models.PositiveIntegerField(blank=True, default=None, help_text='See https://psutil.readthedocs.io/en/latest/', null=True)),
                ('site_user', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to=settings.AUTH_USER_MODEL)),
                ('dataset', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='statuses', to='crunch.dataset')),
            ],
            options={
                'verbose_name_plural': 'statuses',
            },
        ),
        migrations.AddField(
            model_name='attribute',
            name='dataset',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attributes', to='crunch.dataset'),
        ),
    ]



"""
Django settings for django-crunch project.

Generated by 'django-admin startproject' using Django 4.0.2.

For more information on this file, see
https://docs.djangoproject.com/en/4.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.0/ref/settings/
"""
import os
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path.cwd()

PROJ_DIR = Path(__file__).resolve().parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'django-insecure-ibbp#2t@pqvvt+!e-qn2&y*oq)@(^!rs4l424uf@0n970xic%d'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'djangocms_admin_style',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'polymorphic_tree',
    'polymorphic',
    "django_extensions",
    'rest_framework',
    "rest_framework.authtoken",
    'django.contrib.sites',
    'easy_thumbnails',

    'filer',
    'mptt',

    'cms',
    'sekizai',
    'menus',
    'treebeard',
    'cms_bootstrap',
    "djangocms_themata",

    'djangocms_text_ckeditor',
    'djangocms_icon',
    'djangocms_link',
    'djangocms_file',
    'djangocms_picture',
    'djangocms_video',
    'djangocms_googlemap',
    'djangocms_snippet',
    'djangocms_style',
    'meta',
    'djangocms_page_meta',

    'djangocms_bootstrap4',
    'djangocms_bootstrap4.contrib.bootstrap4_alerts',
    'djangocms_bootstrap4.contrib.bootstrap4_badge',
    'djangocms_bootstrap4.contrib.bootstrap4_card',
    'djangocms_bootstrap4.contrib.bootstrap4_carousel',
    'djangocms_bootstrap4.contrib.bootstrap4_collapse',
    'djangocms_bootstrap4.contrib.bootstrap4_content',
    'djangocms_bootstrap4.contrib.bootstrap4_grid',
    'djangocms_bootstrap4.contrib.bootstrap4_jumbotron',
    'djangocms_bootstrap4.contrib.bootstrap4_link',
    # 'djangocms_bootstrap4.contrib.bootstrap4_listgroup',
    # 'djangocms_bootstrap4.contrib.bootstrap4_media',
    # 'djangocms_bootstrap4.contrib.bootstrap4_picture',
    'djangocms_bootstrap4.contrib.bootstrap4_tabs',
    # 'djangocms_bootstrap4.contrib.bootstrap4_utilities',

    'crunch',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',

    "django.middleware.locale.LocaleMiddleware",

    'cms.middleware.user.CurrentUserMiddleware',
    'cms.middleware.page.CurrentPageMiddleware',
    'cms.middleware.toolbar.ToolbarMiddleware',
    'cms.middleware.language.LanguageCookieMiddleware',

]

ROOT_URLCONF = 'crunch.django.proj.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'cms.context_processors.cms_settings',
                'sekizai.context_processors.sekizai',
            ],
        },
    },
]

WSGI_APPLICATION = 'crunch.django.proj.wsgi.application'


# Database
# https://docs.djangoproject.com/en/4.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}


# Password validation
# https://docs.djangoproject.com/en/4.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/

LANGUAGE_CODE = 'en'
LANGUAGES = [
    ('en', 'English'),
]


TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.0/howto/static-files/

STATIC_URL = 'static/'
STATIC_ROOT = str(PROJ_DIR / "staticfiles")

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        'rest_framework.authentication.BasicAuthentication',
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.TokenAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": ("rest_framework.permissions.DjangoModelPermissions",),
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.PageNumberPagination',
    'PAGE_SIZE': 20,
}

SITE_ID = 1
X_FRAME_OPTIONS = 'SAMEORIGIN'

CMSPLUGIN_CASCADE_PLUGINS = []
CMS_TOOLBAR_ANONYMOUS_ON = False

CMS_PERMISSION = True
CMS_TEMPLATES = [
    ('base.html', 'Base Template'),
]

MEDIA_ROOT = os.path.join(BASE_DIR, 'storage/')
MEDIA_URL = '/storage/'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crunch',
 'crunch.client',
 'crunch.django',
 'crunch.django.app',
 'crunch.django.proj',
 'crunch.management.commands',
 'crunch.migrations',
 'crunch.templatetags']

package_data = \
{'': ['*'], 'crunch': ['templates/*', 'templates/crunch/*']}

install_requires = \
['Django>=3.2,<4.0',
 'anytree>=2.8.0,<3.0.0',
 'django-cms>=3.9.0,<4.0.0',
 'django-extensions>=3.1.5,<4.0.0',
 'django-mptt>=0.13.4,<0.14.0',
 'django-next-prev>=1.1.0,<2.0.0',
 'django-polymorphic-tree>=2.1,<3.0',
 'django-polymorphic>=2,<3.1',
 'django-storages[boto3]>=1.12.2,<2.0.0',
 'djangocms-admin-style>=2.0.2,<3.0.0',
 'djangocms-attributes-field>=2.0.0,<3.0.0',
 'djangocms-bootstrap4>=2.0.0,<3.0.0',
 'djangocms-bootstrap>=1.1.2,<2.0.0',
 'djangocms-file>=3.0.0,<4.0.0',
 'djangocms-googlemap>=2.0.0,<3.0.0',
 'djangocms-link>=3.0.0,<4.0.0',
 'djangocms-page-meta>=1.0.1,<2.0.0',
 'djangocms-picture>=3.0.0,<4.0.0',
 'djangocms-snippet>=3.0.0,<4.0.0',
 'djangocms-style>=3.0.0,<4.0.0',
 'djangocms-text-ckeditor>=4.0.0,<5.0.0',
 'djangocms-themata>=0.1.0,<0.2.0',
 'djangocms-video>=3.0.0,<4.0.0',
 'djangorestframework>=3.13.1,<4.0.0',
 'humanize>=4.0.0,<5.0.0',
 'ipykernel>=6.9.1,<7.0.0',
 'pandas>=1.4.4,<2.0.0',
 'pydeck>=0.7.1,<0.8.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'rich>=11.1.0,<12.0.0',
 'setuptools>=65.3.0,<66.0.0',
 'snakemake>=6.15.5,<7.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer[all]>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['crunch = crunch.client.main:app',
                     'crunchsite = crunch.django.proj.manage:main']}

setup_kwargs = {
    'name': 'django-crunch',
    'version': '0.1.12',
    'description': 'A data processing orcestration tool.',
    'long_description': '================================================================\ndjango-crunch\n================================================================\n\n.. image:: https://raw.githubusercontent.com/rbturnbull/django-crunch/main/docs/images/crunch-banner.svg\n\n.. start-badges\n\n|testing badge| |coverage badge| |docs badge| |black badge|\n\n.. |testing badge| image:: https://github.com/rbturnbull/django-crunch/actions/workflows/testing.yml/badge.svg\n    :target: https://github.com/rbturnbull/django-crunch/actions\n\n.. |docs badge| image:: https://github.com/rbturnbull/django-crunch/actions/workflows/docs.yml/badge.svg\n    :target: https://rbturnbull.github.io/django-crunch\n    \n.. |black badge| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    \n.. |coverage badge| image:: https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/rbturnbull/d83b00666fad82df59a814083a09d1c1/raw/coverage-badge.json\n    :target: https://rbturnbull.github.io/django-crunch/coverage/\n    \n.. end-badges\n\n\n.. start-quickstart\n\nA data processing orcestration tool.\nCrunch allows you to visualize the datasets, orchestrate and manage the processing online and present the results to the world.\n\nDescription\n===========\n\nCrunch coordinates three components. First there is a web application in the cloud. \nCrunch includes a Django app which can be included in a website built using the Django framework for building database driven websites. \nThe website allows users to create what we call ``datasets``. \nEach dataset corresponds to one collection of files and metadata which is designed to be run through the workflow. \nEach dataset has its own page on the website which displays all the information about it.\n\nThe second component is data storage. \nThe main use-case for crunch is when your datasets are so many and so large that you cannott fit them all on to the disk where you are doing your computation. \nWith crunch, the data can be stored in any of the media storage options available with Django. \nThese can be Amazon S3, Google Cloud or many other storage options. \n\nThe third component is the client which runs at the place where you are performing the computation. \nThis could be in a high-performance computing environment, or it could be on a virtual machine in the cloud or it could be just on your own laptop. \nThe user just runs the crunch command line tool. \nThe command line tool communicates with the website to find out which dataset should be processed next, \nthen it copies it from storage and saves it locally. \nThen it processes the dataset using a pre-defined workflow provided by the user. \nWhen it is finished, it copies back the data to the storage with the resulting files. \nAt each stage the user can see the status on the website interface. \nIf you run the `crunch loop`` command then the client continually loops through the datasets until they are completely finished. \nYou can run as many clients in parallel as you have computing resources.\n\nOnce each dataset is processed, the resulting files can be accessed via the website. \nThe permissions for the website can be set dynamically so that users can restrict access \nto just the team for while the data is being processed and once the results are ready for the world then you can allow access to the public.\n\n\nInstallation\n==================================\n\nThe crunch app for a Django website and the command-line client are installed with pip:\n\n.. code-block:: bash\n\n    pip install git+https://github.com/rbturnbull/django-crunch\n\n\nInstall the crunch app to the Djanco website project by adding it to the settings:\n\n.. code-block:: python\n\n    INSTALLED_APPS += [\n        "crunch",\n    ]\n\nThen add the urls to your main urls.py:\n\n.. code-block:: python\n\n    urlpatterns += [\n        path(\'crunch/\', include(\'crunch.django.app.urls\'))),    \n    ]\n\nThe path ``crunch/`` can be changed to be whatever you choose.\n\nUsage\n==================================\n\nCreate projects, datasets, items and attributes on the website using the HTML interface, the crunch command-line client or the Python API.\n\nUpload initial data for each dataset as needed to the storage using the Crunch HTML interface or direct to the folder for each dataset on the storage.\n\nThen process each dataset at the location where you are performing your compute with the crunch client. All datasets can be processed with the single command:\n\n.. code-block:: bash\n\n    crunch loop\n\n.. end-quickstart\n\nCredits\n==================================\n\n.. start-credits\n\nRobert Turnbull, Mar Quiroga and Simon Mutch from the Melbourne Data Analytics Platform.\n\nPublication and citation details to follow.\n\n.. end-credits\n',
    'author': 'Robert Turnbull, Mar Quiroga and Simon Mutch',
    'author_email': 'robert.turnbull@unimelb.edu.au',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rbturnbull/django-crunch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<=3.11',
}


setup(**setup_kwargs)



import platform, socket, re, uuid, psutil, os
import getpass

import os
from pathlib import Path
import subprocess

def version() -> str:
    """
    Gets the version number of the django-crunch module.

    Returns:
        str: The current version.
    """
    import importlib.metadata
    version = importlib.metadata.version("django-crunch")
    return version


# Return the git revision as a string
def git_revision(directory:Path=None) -> str:
    """
    Gets the git revision hash for a directory.

    Adapted from https://stackoverflow.com/a/40170206 which was taken from NumPy.

    Args:
        directory (Path, optional): The directory we are interested in. Defaults to None in which case it uses the directory of the current source file.

    Returns:
        str: The git hash for the current revision.
    """
    directory = directory or Path(__file__).parent
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout = subprocess.PIPE, env=env, cwd=directory).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION

def get_diagnostic(diagnostics:dict, key:str, func, default=""):
    try:
        result = func()
    except Exception:
        result = default
    diagnostics[key] = result
    return result

def get_diagnostics() -> dict:
    """
    Gets diagnostic information about the current environment.

    Used when sending status updates to a crunch hosted site.

    Returns:
        dict: A dictionary with the diagnostic information.
    """
    diagnostics = dict()

    # adapts code from here: https://stackoverflow.com/questions/3103178/how-to-get-the-system-info-with-python
    get_diagnostic(diagnostics, 'agent_user', lambda: getpass.getuser())
    get_diagnostic(diagnostics, 'version', lambda: version())
    get_diagnostic(diagnostics, 'revision', lambda: git_revision())
    get_diagnostic(diagnostics, 'system', lambda: platform.system())
    get_diagnostic(diagnostics, 'system_release', lambda: platform.release())
    get_diagnostic(diagnostics, 'system_version', lambda: platform.version())
    get_diagnostic(diagnostics, 'machine', lambda: platform.machine())
    get_diagnostic(diagnostics, 'hostname', lambda: socket.gethostname())
    get_diagnostic(diagnostics, 'ip_address', lambda: socket.gethostbyname(socket.gethostname())) # this could use the hostname from above
    get_diagnostic(diagnostics, 'mac_address', lambda: ':'.join(re.findall('..', '%012x' % uuid.getnode())))
    get_diagnostic(diagnostics, 'memory_total', lambda: psutil.virtual_memory().total, default=None)
    get_diagnostic(diagnostics, 'memory_free', lambda: psutil.virtual_memory().available, default=None)
    get_diagnostic(diagnostics, 'disk_total', lambda: psutil.disk_usage(os.getcwd()).total, default=None)
    get_diagnostic(diagnostics, 'disk_free', lambda: psutil.disk_usage(os.getcwd()).free, default=None)

    return diagnostics


from operator import mod
from typing import List
import re
from typing import Type
from django.db import models
from django_extensions.db.fields import AutoSlugField
from django.utils.text import slugify
from django.urls import reverse
from django.contrib.auth import get_user_model
from django.utils.html import format_html
from django.db.models import OuterRef, Subquery
from mptt.models import MPTTModel, TreeForeignKey
import humanize
from polymorphic.models import PolymorphicModel
from django_extensions.db.models import TimeStampedModel
from next_prev import next_in_order, prev_in_order
from polymorphic_tree.models import PolymorphicMPTTModel, PolymorphicTreeForeignKey


from . import enums, storages

User = get_user_model()


def OptionalCharField(max_length=255, default="", blank=True, **kwargs):
    return models.CharField(
        max_length=max_length, default=default, blank=blank, **kwargs
    )


class NextPrevMixin(models.Model):
    class Meta:
        abstract = True

    def next_in_order(self, **kwargs):
        return next_in_order(self, **kwargs)

    def prev_in_order(self, **kwargs):
        return prev_in_order(self, **kwargs)

    def get_admin_url(self):
        return reverse(
            f"admin:{self._meta.app_label}_{self._meta.model_name}_change",
            args=(self.pk,),
        )


class Item(NextPrevMixin, TimeStampedModel, PolymorphicMPTTModel):
    """
    A general class of object which can be placed in a hierachical tree. 

    Each item can be viewed online with a details page and each item can be given any number of attributes to store metadata.
    """
    parent = PolymorphicTreeForeignKey(
        "self",
        blank=True,
        null=True,
        default=None,
        related_name="children",
        on_delete=models.SET_DEFAULT,
    )
    name = models.CharField(max_length=1023, unique=True)
    description = models.CharField(
        max_length=1023,
        default="",
        blank=True,
        help_text="A short description in a sentence or more of this item.",
    )
    details = models.TextField(
        default="",
        blank=True,
        help_text="A detailed description of this item (written in Markdown).",
    )
    slug = AutoSlugField(populate_from="name", unique=True, max_length=255)
    # TODO Add tags

    def slugify_function(self, content):
        slug = slugify(content)
        if self.parent:
            return f"{self.parent.slug}:{slug}"
        return slug

    class Meta(PolymorphicMPTTModel.Meta):
        unique_together = ("parent", "slug")
        ordering = ('created', 'pk')

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("crunch:item-detail", kwargs={"slug": self.slug})

    def items(self):
        return self.get_children()

    def descendant_attributes(
        self, attribute_type: Type = None, include_self: bool = True
    ) -> models.QuerySet:
        """Returns a queryset with all the attributes of the descendants of this item.

        Args:
            attribute_type (Type, optional): The type of the attribute to filter for. If `None` then it uses the `Attribute` class.
            include_self (bool, optional): Whether or not to include attributes of this item. Defaults to True.

        Returns:
            models.QuerySet: A queryset of attributes of items descended from this item.
        """
        attribute_type = attribute_type or Attribute
        return attribute_type.objects.filter(
            item__in=self.get_descendants(include_self=include_self)
        )

    def descendant_total_filesize(self) -> int:
        """
        Sums all the filesize attributes for this item and its descendants.

        Returns:
            int: The total sum of the filesize attributes of this item and all its descendants. 
                If there are no filesize attributes then it returns None.
        """
        
        filesize_attributes = self.descendant_attributes(
            attribute_type=FilesizeAttribute, include_self=True
        )

        if not filesize_attributes:
            return None

        return filesize_attributes.aggregate(models.Sum("value"))["value__sum"]

    def descendant_total_filesize_readable(self) -> str:
        """
        Sums all the filesize attributes for this item and its descendants and converts it to a human readable string.

        Returns:
            str: The total sum of the filesize attributes of this item and all its descendants in a human readable string. 
                If there are no filesize attributes then it returns 'None'.
        """
        descendant_total_filesize = self.descendant_total_filesize()

        if descendant_total_filesize:
            return humanize.naturalsize(descendant_total_filesize)

        return "None"

    def map(self):
        from .mapping import item_map

        return item_map(self)

    def descendant_latlongattributes(self):
        return self.descendant_attributes(
            attribute_type=LatLongAttribute, include_self=True
        )

    def has_descendant_latlongattributes(self):
        return self.descendant_latlongattributes().count() > 0

    def reslugify_descendants(self):
        for item in self.get_descendants(include_self=True):
            item.slug = item.slugify_function(item.name)
            item.save()


class Project(Item):
    """ 
    An item which collects a number of datasets which should be run with the same workflow. 

    Projects ought not to have parents in the tree structure of items.
    """
    workflow = models.TextField(
        default="",
        blank=True,
        help_text="URL to snakemake repository/shell script or its content.",
    )
    # More workflow languages need to be supported.
    # TODO assert parent is none

    def get_absolute_url(self):
        return reverse("crunch:project-detail", kwargs={"slug": self.slug})

    def unprocessed_datasets(self) -> models.QuerySet:
        """
        Returns a QuerySet of all datasets in this project that are not complete and are not locked.
        """
        return Dataset.unprocessed().filter(id__in=self.items())

    def completed_datasets(self) -> models.QuerySet:
        """
        Returns a QuerySet of all datasets in this project that are completed.
        """
        return Dataset.completed().filter(id__in=self.items())

    def running_datasets(self) -> models.QuerySet:
        """
        Returns a QuerySet of all datasets in this project that are running.
        """
        return Dataset.running().filter(id__in=self.items())

    def failed_datasets(self) -> models.QuerySet:
        """
        Returns a QuerySet of all datasets in this project that have failed.
        """
        return Dataset.failed().filter(id__in=self.items())

    def next_unprocessed_dataset(self) -> "Dataset":
        return self.unprocessed_datasets().first()


class Dataset(Item):
    """ 
    An item should be run once in a workflow.

    The parent of a dataset should be its project.
    """
    base_file_path = models.CharField(max_length=4096, default="", blank=True)
    locked = models.BooleanField(
        default=False,
        help_text="If the dataset is locked then it will not show up in the loop of available datasets.",
    )

    def save(self, *args, **kwargs):
        assert isinstance(self.parent, Project)

        if not self.base_file_path:
            self.base_file_path = storages.default_dataset_path(self.parent.slug, self.slug)
        return super().save(*args, **kwargs)

    def get_absolute_url(self) -> str:
        return f"{self.parent.get_absolute_url()}datasets/{self.slug}"

    @classmethod
    def completed_ids(cls) -> List[int]:
        """
        Returns a list of ids of all datasets which have a status with stage UPLOAD and state SUCCESS.
        """
        return Status.completed().values_list("dataset__id", flat=True)

    @classmethod
    def completed(cls) -> models.QuerySet:
        """
        Returns a QuerySet of all datasets which have a status with stage UPLOAD and state SUCCESS.
        """
        return cls.objects.filter(id__in=cls.completed_ids())

    @classmethod
    def incomplete(cls) -> models.QuerySet:
        """
        Returns a QuerySet of all datasets that are not complete (including unprocessed, running and failed datasets).
        """
        return cls.objects.exclude(id__in=cls.completed_ids())

    @classmethod
    def unprocessed(cls) -> models.QuerySet:
        """
        Returns a QuerySet of all incomplete datasets that are not locked.
        """
        return cls.incomplete().filter(locked=False)

    @classmethod
    def inprocess(cls) -> models.QuerySet:
        """
        Returns a QuerySet of all incomplete datasets that are locked (including running and failed datasets).
        """
        return cls.incomplete().filter(locked=True)

    @classmethod
    def has_status(cls) -> models.QuerySet:
        """
        Returns a QuerySet of all datasets with at least one status.
        """
        statuses = Status.objects.filter(dataset=OuterRef("pk"))
        return cls.objects.filter(models.Exists(statuses))

    @classmethod
    def failed(cls) -> models.QuerySet:
        """
        Returns a QuerySet of all incomplete unlocked datasets where the latest status has a state of 'FAILED'.
        """
        newest_statuses = Status.objects.filter(dataset=OuterRef("pk")).order_by(
            "-created"
        )
        annotated = cls.inprocess().annotate(
            newest_status_state=Subquery(newest_statuses.values("state")[:1])
        )
        return annotated.filter(newest_status_state=enums.State.FAIL)

    @classmethod
    def running(cls) -> models.QuerySet:
        """
        Returns a QuerySet of all incomplete unlocked datasets where the latest status does not have a state of 'FAILED'.
        """
        return cls.inprocess().exclude(id__in=cls.failed())

    @classmethod
    def next_unprocessed(cls) -> "Dataset":
        return cls.unprocessed().first()

    def files(self):
        return storages.storage_walk(self.base_file_path)


class Status(NextPrevMixin, TimeStampedModel):
    dataset = models.ForeignKey(
        Dataset, on_delete=models.CASCADE, related_name="statuses"
    )
    site_user = models.ForeignKey(
        User, on_delete=models.SET_DEFAULT, default=None, blank=True, null=True
    )
    stage = models.IntegerField(choices=enums.Stage.choices)
    state = models.IntegerField(choices=enums.State.choices)
    note = models.TextField(default="", blank=True)
    # Diagnostic info
    agent_user = OptionalCharField(
        help_text="The name of the user running the agent (see https://docs.python.org/3/library/getpass.html)."
    )
    version = OptionalCharField(
        help_text="The django-crunch version number of the agent."
    )
    revision = OptionalCharField(
        help_text="The django-crunch git revision hash of the agent."
    )
    # terminal = OptionalCharField(help_text="the tty or pseudo-tty associated with the agent user (see https://psutil.readthedocs.io/en/latest/).")
    system = OptionalCharField(
        help_text="Returns the system/OS name, such as 'Linux', 'Darwin', 'Java', 'Windows' (see https://docs.python.org/3/library/platform.html)."
    )
    system_release = OptionalCharField(
        help_text="Returns the system’s release, e.g. '2.2.0' or 'NT' (see https://docs.python.org/3/library/platform.html)."
    )
    system_version = OptionalCharField(
        help_text="Returns the system’s release version, e.g. '#3 on degas' (see https://docs.python.org/3/library/platform.html)."
    )
    machine = OptionalCharField(
        help_text="Returns the machine type, e.g. 'i386' (see https://docs.python.org/3/library/platform.html)."
    )
    hostname = OptionalCharField(
        help_text="The hostname of the machine where the agent was running (see https://docs.python.org/3/library/socket.html)."
    )
    ip_address = OptionalCharField(
        help_text="The hostname in IPv4 address format (see https://docs.python.org/3/library/socket.html)."
    )
    mac_address = OptionalCharField(
        help_text="The hardware address  (see https://docs.python.org/3/library/uuid.html)."
    )
    memory_total = models.BigIntegerField(
        default=None,
        blank=True,
        null=True,
        help_text="See https://psutil.readthedocs.io/en/latest/",
    )
    memory_free = models.BigIntegerField(
        default=None,
        blank=True,
        null=True,
        help_text="See https://psutil.readthedocs.io/en/latest/",
    )
    disk_total = models.BigIntegerField(
        default=None,
        blank=True,
        null=True,
        help_text="See https://psutil.readthedocs.io/en/latest/",
    )
    disk_free = models.BigIntegerField(
        default=None,
        blank=True,
        null=True,
        help_text="See https://psutil.readthedocs.io/en/latest/",
    )

    class Meta:
        verbose_name_plural = "statuses"

    def __str__(self):
        return f"{self.dataset}: {self.get_stage_display()} {self.get_state_display()}"

    def save(self, *args, **kwargs):

        # Lock dataset if necessary
        assert isinstance(self.dataset, Dataset)
        if not self.dataset.locked:
            self.dataset.locked = True
            self.dataset.save()

        super().save(*args, **kwargs)

    @classmethod
    def completed(cls):
        return Status.objects.filter(
            stage=enums.Stage.UPLOAD, state=enums.State.SUCCESS
        )


class Attribute(NextPrevMixin, TimeStampedModel, PolymorphicModel):
    item = models.ForeignKey(Item, on_delete=models.CASCADE, related_name="attributes")
    key = models.CharField(max_length=255)

    def value_dict(self):
        return dict(key=self.key)

    def value_str(self):
        raise NotImplementedError("value_str not implemented for this attribute class")

    def value_html(self):
        return self.value_str()

    def __str__(self):
        return f"{self.key}: {self.value_str()}"

    def type_str(self) -> str:
        """
        Returns a string describing this type of attribute.

        By default it returns the class name with spaces added where implied by camel case.

        Returns:
            str: The type of this attribute as a string.
        """
        class_name = self.__class__.__name__
        if class_name.endswith("Attribute"):
            class_name = class_name[: -len("Attribute")]

        return re.sub(r"((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))", r" \1", class_name)


class ValueAttribute(Attribute):
    # Child classes need to give a 'value' field.

    class Meta:
        abstract = True

    def value_dict(self):
        d = super().value_dict()
        d["value"] = self.value
        return d

    def value_str(self):
        return self.value


class CharAttribute(ValueAttribute):
    """ An attribute for storing metadata as a string (of maximum length 1023 characters). """
    value = models.CharField(max_length=1023)


class FloatAttribute(ValueAttribute):
    value = models.FloatField()


class IntegerAttribute(ValueAttribute):
    value = models.IntegerField()


class FilesizeAttribute(ValueAttribute):
    value = models.PositiveBigIntegerField(
        help_text="The filesize of this item in bytes."
    )

    def value_str(self):
        return humanize.naturalsize(self.value)


class BooleanAttribute(ValueAttribute):
    value = models.BooleanField()


class DateTimeAttribute(ValueAttribute):
    value = models.DateTimeField()


class DateAttribute(ValueAttribute):
    value = models.DateField()


class URLAttribute(ValueAttribute):
    value = models.URLField(max_length=1023)

    def value_html(self):
        return format_html(
            "<a href='{}'>{}</a>",
            self.value,
            self.value,
        )


class LatLongAttribute(Attribute):
    """ An attribute for storing a geolocation (in decimal degrees). """
    latitude = models.DecimalField(
        max_digits=12,
        decimal_places=9,
        help_text="The latitude of this location in decimal degrees.",
    )
    longitude = models.DecimalField(
        max_digits=12,
        decimal_places=9,
        help_text="The longitude of this location in decimal degrees.",
    )

    def value_dict(self):
        d = super().value_dict()
        d["latitude"] = self.latitude
        d["longitude"] = self.longitude
        return d

    def value_str(self):
        return f"{self.latitude:+}{self.longitude:+}/"

    def value_html(self):
        return format_html(
            "<a href='https://www.google.com/maps/place/{},{}'>{}</a>",
            self.latitude,
            self.longitude,
            self.value_str(),
        )



# Generated by Django 3.2.15 on 2022-08-11 02:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crunch', '0008_auto_20220811_1223'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attribute',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='dataset',
            name='base_file_path',
            field=models.CharField(blank=True, default='', max_length=4096),
        ),
        migrations.AlterField(
            model_name='item',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='status',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]
